﻿<#
.SYNOPSIS
Gets regular, detailed statistics for VMs and hosts, and saves them in CSV files.

.DESCRIPTION
Statistics are collected through vCenter. 

Metrics to retrieve are hardcoded in the script as constants.

The script checks for a 'last-collection.time' to see when it has been
last run, otherwise collects the stats from the last 30 minutes.

Implements the -Verbose common parameter to display verbose messages.

.PARAMETER vmNames 
The virtual machines to collect statistics for.

.PARAMETER hostNames
The ESXi hosts to collect statistics for.

.PARAMETER csvFolder
The folder the CSV files are saved into.

.EXAMPLE
Get-VMwareStat -vmNames vm1,vm2 -hostNames host1 -csvFolder c:\stat

.NOTES
Author: Micskei Zoltán, 2012.08.26.

#>

[CmdletBinding()]
param(
	[Parameter(Mandatory=$false)][string[]] $vmNames,
	[Parameter(Mandatory=$false)][string[]] $hostNames,
	[Parameter(Mandatory=$true)][string] $csvFolder	
)

#----------- Global variables ---------------

$VCENTER_NAME = "vcenter.example.local"
$LAST_COLLECTION_FILE_NAME = "last-collection.time"

# metrics to collect from VMs (for others see the VI API SDK)
$VM_METRIC_NAMES = "cpu.idle.summation", "cpu.ready.summation", "cpu.run.summation", "cpu.swapwait.summation", "cpu.system.summation",
    "cpu.usage.average", "cpu.usagemhz.average", "cpu.used.summation", "cpu.wait.summation", 
	"mem.active.average", "mem.compressed.average", "mem.consumed.average", "mem.granted.average", "mem.overhead.average", "mem.shared.average", 
	"mem.swapinrate.average", "mem.swapoutrate.average", "mem.swaptarget.average", "mem.vmmemctl.average", "mem.vmmemctltarget.average", "mem.zero.average",
    "virtualdisk.read.average", "virtualdisk.totalreadlatency.average", "virtualdisk.write.average", "virtualdisk.totalwritelatency.average",
	"net.received.average","net.transmitted.average",
    "sys.uptime.latest"

# metrics to collect from hosts
$HOST_METRIC_NAMES = "cpu.usage.average", "cpu.usagemhz.average", "cpu.used.summation", 
    "mem.active.average", "mem.consumed.average", "mem.granted.average", "mem.overhead.average", "mem.shared.average", "mem.sharedcommon.average",
	"mem.state.latest", "mem.sysUsage.average", "mem.swapinRate.average", "mem.swapoutRate.average", 
    "mem.swapused.average", "mem.usage.average", "mem.vmmemctl.average", "mem.zero.average",
    "disk.commands.summation", "disk.commandsAborted.summation",
    "disk.deviceLatency.average", "disk.kernelLatency.average", "disk.queueLatency.average", 
    "disk.totallatency.average", "disk.totalreadlatency.average", "disk.totalwritelatency.average",
    "disk.numberRead.summation", "disk.numberWrite.summation", "disk.read.average", "disk.write.average",  
    "storageadapter.totalreadlatency.average", "storageadapter.totalwritelatency.average", "storagepath.totalreadlatency.average", "storagepath.totalwritelatency.average",
    "net.broadcastrx.summation", "net.broadcasttx.summation", "net.multicastrx.summation", "net.multicasttx.summation",
    "net.droppedRx.summation", "net.droppedTx.summation", "net.packetsRx.summation", "net.packetsTx.summation", "net.received.average", "net.transmitted.average",
    "sys.uptime.latest"

# Stop on all errors
$ErrorActionPreference = "Stop"

#----------- Functions ---------------

function Initialize-PowerCLI
{
	# Check if PowerCLI core snapin is loaded
	if ( (Get-PSSnapin | ? {$_.Name -like "VMware*"}) -eq $null )
	{
		Add-PSSnapin VMware.VimAutomation.Core
	}
	
	Connect-VIServer $VCENTER_NAME > $null
    Write-Verbose "Connected to $VCENTER_NAME"
}

function Collect-Statistics
{
	param(
		[string] $entityName,
		[string[]] $METRIC_NAMES,
        [datetime] $collectionStart,
        [datetime] $collectionFinish			
	)	
	
	$csvName = "$($entityName)-$(Get-Date -uFormat '%Y%m%d').csv"
	$csvPath = Join-Path $csvFolder $csvName

	# CSV header is dynamic because of multi-instance metrics (e.g. cpu.* has as many instances as the VM's VCPU),
	# thus we can write it later when at least one sample is queried.
	$writeCsvHeader = $false
	if (! ( Test-Path $csvPath))
	{
		$writeCsvHeader = $true
	}
	
	$stats = Get-Stat -Entity $entityName -Realtime -Stat $METRIC_NAMES -Start $collectionStart -Finish $collectionFinish 
	
    # group the stats by timestamps
    $groupedStats = $stats | Group-Object -Property timestamp -AsHashTable
    
    # loop through all the distinct timestamps
    $groupedStats.GetEnumerator() | Sort-Object -Property name | % {				
		$csvHeader = "timestamp;"
		$out = "$($_.name);"

        # group the values for the actual timestamp by metrics
        $valuesByMetric = $_.value | Group-Object -Property MetricId -AsHashTable

		# need a for loop, because the iterator's order is not fixed
		for ($i = 0; $i -lt $METRIC_NAMES.Length; $i++)
		{
			# get all the different instances
			$instances = $valuesByMetric[$METRIC_NAMES[$i]] | select instance -Unique | sort instance
			
			foreach ($instance in $instances)
			{			
				$out += ($valuesByMetric[$METRIC_NAMES[$i]] | ? { ($_.Instance -eq $instance.instance) }).Value
				$out += ";"
				
				if ($writeCsvHeader)
				{
					$csvHeader += "$($METRIC_NAMES[$i])_$($instance.instance);"
				}					
			}
		}

		if ($writeCsvHeader)
		{
			Out-File -FilePath $csvPath -InputObject $csvHeader -Encoding utf8
			$writeCsvHeader = $false
		}
		
        Out-File -FilePath $csvPath -InputObject $out -Encoding utf8 -Append

#Write-Verbose "$((Get-Date).Second).$((Get-Date).Millisecond) one time stamp finished"
	}
}

#----------- Main code of script ---------------

if (! (Test-Path $csvFolder))
{
	throw "CSV folder does not exists: $csvFolder"
}

Initialize-PowerCLI

# interval to collect statistics
# (ESXi keeps realtime data for 1 hour, and 30 minutes can be reasonably processed)
$finish = Get-Date
$start = $finish.AddMinutes(-30)

# was there valid last run, which can be used as Start for Get-Stat
if (Test-Path $LAST_COLLECTION_FILE_NAME)
{
    $content = Get-Content $LAST_COLLECTION_FILE_NAME

    try
    {
        $d = [datetime] $content
        $start = $d
        Write-Verbose "Start date reloaded from file $LAST_COLLECTION_FILE_NAME"
    }
    catch
    {
        # do nothing, if in the file there was no valid datetime
        # use the previosly set value
        Write-Verbose "Start date could not be reloaded from file $LAST_COLLECTION_FILE_NAME"
    }
}

Write-Verbose "Collection  start: $start"
Write-Verbose "Collection finish: $finish"

Write-Verbose "$((Get-Date).ToString()) Starting to collect statistics"

if (! ($vmNames -eq $null))
{
    $vmNames | % {
	    $vmName = $_.Trim()
	
	    Write-Verbose "Processing VM $vmName"
	
	    Collect-Statistics $vmName $VM_METRIC_NAMES $start $finish
    }
}

if (! ($hostNames -eq $null))
{
    $hostNames | % {
	    $hostName = $_.Trim()
	
	    Write-Verbose "Processing host $hostName"
	
	    Collect-Statistics $hostName $HOST_METRIC_NAMES  $start $finish
    }
}

# save last successful run
Write-Output $finish.ToString() > $LAST_COLLECTION_FILE_NAME

Write-Verbose "$((Get-Date).ToString()) Finished collecting data"